/****************************************************************************

  (c) SYSTEC electronic GmbH, D-07973 Greiz, August-Bebel-Str. 29
      www.systec-electronic.de

  Project:      PLCmodule C14

  Description:  Serial driver for C16x with external UART TL16C550

  -------------------------------------------------------------------------

                $RCSfile: TL16C550.c,v $
                
                $Author: A.Suess $

                $Revision: 1.2 $  $Date: 2004/10/06 12:12:51 $
                
                $State: Exp $
                
                Build Environment:
                KEIL uVision 2

  -------------------------------------------------------------------------

  Revision History:
   
  2003/07/07 -avc:   Start der Implementierung

  2003/10/24 -as:    Implementierung berarbeitet:
                     - Semaphoren werden nicht mehr bentigt
                     - Die Ringpuffer-Verwaltung kommt ohne Interruptsperre
                       aus.
                     - Fehler in der Funktion SetState behoben
                     - Lschen des Sende- und Empfangspuffers mit Hilfe
                       von SetState implementiert
                     - XON-XOFF-Protokoll implementiert

  2003/10/27  -as:   Treiber fr UART TL16C550 erstellt

  2003/11/12  -rs:   Anpassung an XC161, Untersttzung von UART1 und UART2,
                     direkte Interruptanbindung als Alternative zu RTOS

  2004/10/06  -as:   Fehler bei Parametrierung der Paritt behoben
                     
****************************************************************************/

#include "generic/global.h"
#ifdef _USE_RTOS_
    #include "rtos/rtosx16x.h"
#endif
#include "../include/SioDrv.h"


#ifdef __C166__
    #include <xc161.h>
    #include <intrins.h>
#endif 





/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*          G L O B A L   D E F I N I T I O N S                            */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

//---------------------------------------------------------------------------
// Auswahl der zu untersttzenden Schnittstelle
//#define UART_NUM    1
//#define UART_NUM    2

#ifndef UART_NUM
    #define UART_NUM  1
#endif



//---------------------------------------------------------------------------
// Definition der instanziierbaren ffentlichen Funktionsnamen

#if (UART_NUM == 1)

    #define SerInit                 S1_SioInit
    #define SerClose                S1_SioClose
    #define TxFifoPutByte           S1_SioTxFifoPutByte
    #define RxFifoGetByte           S1_SioRxFifoGetByte
    #define TxFifoDataSize          S1_SioTxFifoDataSize
    #define RxFifoDataSize          S1_SioRxFifoDataSize
    #define GetState                S1_SioGetState
    #define SetState                S1_SioSetState

    #define SioInterruptHandler     S1_SioInterruptHandler
    #define SioTxTaskHandler        S1_SioTxTaskHandler


#elif (UART_NUM == 2)

    #define SerInit                 S2_SerInit
    #define SerClose                S2_SerClose
    #define TxFifoPutByte           S2_TxFifoPutByte
    #define RxFifoGetByte           S2_RxFifoGetByte
    #define TxFifoDataSize          S2_TxFifoDataSize
    #define RxFifoDataSize          S2_RxFifoDataSize
    #define GetState                S2_GetState
    #define SetState                S2_SetState

    #define SioInterruptHandler     S2_SioInterruptHandler
    #define SioTxTaskHandler        S2_SioTxTaskHandler

#else

    #error ("Invalid or no UART_NUM defined.")

#endif



//---------------------------------------------------------------------------
// const defines
//---------------------------------------------------------------------------
// Einstellung Gre der Ringpuffer
#define SERIAL_RX_FIFO_SIZE        256                 // Size of RX/TX-FIFO in byte
#define SERIAL_TX_FIFO_SIZE        256                 // Size of RX/TX-FIFO in byte

#if (SERIAL_RX_FIFO_SIZE < 49)
    #error 'Empfangspuffer zu klein'
#endif

// Reserved space in RX-FIFO for handshake buffering in byte (RTS is set to LOW if 
// this limit is reached) (default: 48)
#define SER_RXFIFO_STOP_LEVEL      SERIAL_RX_FIFO_SIZE-48



//---------------------------------------------------------------------------
// Einstellung des untersttzten Protokolle und Mode

// Control lines for hardware handshake supported: If not set to TRUE, then 
// XON-XOFF- or no protocol can be used.
#define HW_HANDSHAKE_SUPPORT       TRUE      
//#define HW_HANDSHAKE_SUPPORT       FALSE

// If USE_BLOCKED_MODE == TRUE, the function TxFifoPutByte is waiting
// until transmission is done.
//#define USE_BLOCKED_MODE           TRUE
#define USE_BLOCKED_MODE           FALSE

// If SIO_ECHO_SUPPORTED == TRUE, the upper layer functions can 
// controll sending of character echo.
#define SIO_ECHO_SUPPORTED         TRUE
//#define SIO_ECHO_SUPPORTED         FALSE



//---------------------------------------------------------------------------
// Konfiguration der Hardware
//---------------------------------------------------------------------------

#define UART_CLOCK                      3686400     // f(osz)=3,686MHz


#if (UART_NUM == 1)

    //------------- UART1 auf 38:8000H --------------------------------------

    // UART1: P2.14 -> CC14IO
    #define UART_BASE_ADDRESS           0x380000L   // UART1 ab Addr 38:0000H
    
    #ifdef _USE_RTOS_

        #define TASK_XUART              ???
        #define TASK_TX_HANDLER         ???

    #else

        #define UART_HW_INT_NUM         0x1E        // IntNum CC14IO
        #define UART_HW_INT_REG_TM      CC1_M3      // Mode Register
        #define UART_HW_INT_MASK_TM     0x0F00      // KanalMaske CC14IO
        #define UART_HW_INT_MODE_TM     0x0100      // IntMode CC14IO (Capture RisingEdge)
        #define UART_HW_INT_REG_IC      CC1_CC14IC  // Interrupt Control Register
        #define UART_HW_INT_BIT_IC_IE   CC1_CC14IC_IE
        #define UART_HW_INT_BIT_IC_IR   CC1_CC14IC_IR
        #define UART_HW_INT_INTLVL      4           // interrupt level
        #define UART_HW_INT_GRPXLVL     0           // ext. group prio (XC16x spec)
        #define UART_HW_INT_GRPLVL      2           // group prio

        // Hilfstask
        #define UART_TX_TASK_NUM        0x11        // IntNum CC1
        #define UART_TX_TASK_REG_IC     CC1_CC1IC   // Interrupt Control Register
        #define UART_TX_TASK_BIT_IC_IE  CC1_CC1IC_IE
        #define UART_TX_TASK_BIT_IC_IR  CC1_CC1IC_IR
        #define UART_TX_TASK_INTLVL     3           // interrupt level
        #define UART_TX_TASK_GRPXLVL    0           // ext. group prio (XC16x spec)
        #define UART_TX_TASK_GRPLVL     2           // group prio

    #endif


#elif (UART_NUM == 2)

    //------------- UART2 auf 39:0000H --------------------------------------

    // UART1: P9.2 -> CC18IO
    #define UART_BASE_ADDRESS           0x390000L   // UART1 ab Addr 39:0000H
    
    #ifdef _USE_RTOS_

        #define TASK_XUART              ???
        #define TASK_TX_HANDLER         ???

    #else

        #define UART_HW_INT_NUM         0x32        // IntNum CC18IO
        #define UART_HW_INT_REG_TM      CC2_M4      // Mode Register
        #define UART_HW_INT_MASK_TM     0x0F00      // KanalMaske CC18IO
        #define UART_HW_INT_MODE_TM     0x0100      // IntMode CC18IO (RisingEdge)
        #define UART_HW_INT_REG_IC      CC2_CC18IC  // Interrupt Control Register
        #define UART_HW_INT_BIT_IC_IE   CC2_CC18IC_IE
        #define UART_HW_INT_BIT_IC_IR   CC2_CC18IC_IR
        #define UART_HW_INT_INTLVL      4           // interrupt level
        #define UART_HW_INT_GRPXLVL     0           // ext. group prio (XC16x spec)
        #define UART_HW_INT_GRPLVL      3           // group prio

        #define UART_TX_TASK_NUM        0x12        // IntNum CC02
        #define UART_TX_TASK_REG_IC     CC1_CC2IC   // Interrupt Control Register
        #define UART_TX_TASK_BIT_IC_IE  CC1_CC2IC_IE
        #define UART_TX_TASK_BIT_IC_IR  CC1_CC2IC_IR
        #define UART_TX_TASK_INTLVL     3           // interrupt level
        #define UART_TX_TASK_GRPXLVL    0           // ext. group prio (XC16x spec)
        #define UART_TX_TASK_GRPLVL     3           // group prio

    #endif


#else

    #error ("Invalid or no UART_NUM defined.")

#endif



//---------------------------------------------------------------------------
//  Definition -> UART-Registern
//  Die Bezeichnung der Register orientiert sich am Datenblatt 
//  der UART TL16C550.
//---------------------------------------------------------------------------

// Definition der Registeradressen bei DLAB = 0
#define ADR_RBR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x00))
#define ADR_THR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x00))
#define ADR_IER  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x01))
#define ADR_IIR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x02)) 
#define ADR_FCR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x02))
#define ADR_LCR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x03))
#define ADR_MCR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x04))
#define ADR_LSR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x05))
#define ADR_MSR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x06))
#define ADR_SCR  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x07))

// Definition der Registeradressen bei DLAB = 1
#define ADR_DLL  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x00))
#define ADR_DLM  ((BYTE xhuge*)(UART_BASE_ADDRESS + 0x01))



//---------------------------------------------------------------------------
//  Definition -> UART-Register-Defaults
//---------------------------------------------------------------------------

// Das FCR-Register steuert die Verwendung der internen FIFO. Der
// DMA-Mode wird nicht untersttzt. FIFI wird benutzt, Receiver und
// Transmitter-FIFO werden zurckgesetzt. Der FIFO-Trigger-Level steht
// auf 1. Andererseits wird beim Empfang von Zeichen erst bei Erreichen
// des definierten Trigger-Levels der Empfangsinterrupt ausgelst. 
#define FCR_REGISTER 0x07       // BIT 0       = "1"    FIFO enable
                                // BIT 1       = "1"    Receive FIFO Reset
                                // BIT 2       = "1"    Transmitter FIFO Reset
                                // BIT 3       = "0"    DMA Mode disabled
                                // BIT 4/5     = "00"   Reserved
                                // BIT 6/7     = "00"   Receive FIFO Triggler Level = 1 Byte

// Um auf die Divisor Latch Register zugreifen zu knnen, ist das Bit DLAB im
// Register LCR zu setzen.
#define DLAB_BIT     0x80       // BIT 7       = "1"    DLAB = 1 -> Devisor Latch aktiv

// Der Rx-und TX- Interrupt werden stets gebraucht. Der Rx-Line-Interrupt
// kann benutzt werden um Fehlerzustnde abzulegen. 
#define IER_REGISTER 0x03       // BIT 0       = "1"    Rx Interrupt disabled
                                // BIT 1       = "1"    Tx Interupt disabled
                                // BIT 2       = "0"    Rx Line Interrupt disabled
                                // BIT 3       = "0"    Modem Interrupt disabled
                                // BIT 4/5/6/7 = "0000" not used

#define MCR_REGISTER 0x00       // BIT 0       = "0"    DTR disabled
                                // BIT 1       = "0"    RTS disabled
                                // BIT 2/3     = "00"   OUT1=0 OUT2=0 (LED=on)
                                // BIT 4       = "0"    Loop disabled
                                // Bit 5       = "0"    AFE off
                                // BIT 6/7     = "00" 	not used

#define MSR_REGISTER 0x00       // cleared


// Da der Sende- bzw. Empfangspuffers aus einem Software-FIFO und
// einem Hardware-FIFO besteht, werden diese beiden Makros bentigt,
// um die HW-FIFOs zu lschen
#define CLR_TX_FIFO() (*ADR_FCR |= 0x04)
#define CLR_RX_FIFO() (*ADR_FCR |= 0x02)

//---------------------------------------------------------------------------

#if(HW_HANDSHAKE_SUPPORT)
    
    //define the control lines which are supported by hardware
    #define RTS_SIGNAL_SUPPORT      TRUE
    #define DTR_SIGNAL_SUPPORT      TRUE
    #define CTS_SIGNAL_SUPPORT      TRUE
    #define DSR_SIGNAL_SUPPORT      TRUE
    #define DCD_SIGNAL_SUPPORT      TRUE
    #define RI_SIGNAL_SUPPORT       TRUE

    //define the port and the direction port to wich the 
    //handshake lines are connected 
    #if(RTS_SIGNAL_SUPPORT)
        #define RTS_BASE      (*ADR_MCR)
        #define RTS_BITMASK   0x02                       // bit mask to set RTS port pin in MCR
        #define RTS_BIT       (RTS_BASE & RTS_BITMASK) 
        #define RTS_STATE     (RTS_BIT?1:0)              // returns 1 if signal is activ
        #define CLR_RTS       (RTS_BASE &= ~RTS_BITMASK)
        #define SET_RTS       (RTS_BASE |= RTS_BITMASK)
    #else
        #define RTS_STATE     SIGNAL_NOT_SUPPORTED // RTS signal not supported
        #define CLR_RTS
        #define SET_RTS
    #endif

    #if(DTR_SIGNAL_SUPPORT)
        #define DTR_BASE      (*ADR_MCR)
        #define DTR_BITMASK   0x01                       // bit mask to set DTR port pin in MCR
        #define DTR_BIT       (DTR_BASE & DTR_BITMASK) 
        #define DTR_STATE     (DTR_BIT?1:0)              // returns 1 if signal is activ
        #define CLR_DTR       (DTR_BASE &= ~DTR_BITMASK)
        #define SET_DTR       (DTR_BASE |= DTR_BITMASK)
    #else
        #define DTR_STATE   SIGNAL_NOT_SUPPORTED  // DTR signal not supported
        #define CLR_DTR
        #define SET_DTR
    #endif

    #if(CTS_SIGNAL_SUPPORT)
        #define CTS_BASE      (*ADR_MSR)
        #define CTS_BITMASK   0x10                       // bit mask to set CTS port pin 
        #define CTS_BIT       (CTS_BASE & CTS_BITMASK) 
        #define CTS_STATE     (CTS_BIT?1:0)              // returns 1 if signal is activ
        #define CLR_CTS
        #define SET_CTS
    #else
        #define CTS_STATE   SIGNAL_NOT_SUPPORTED // CTS signal not supported
        #define CLR_CTS
        #define SET_CTS
    #endif
    
    #if(DSR_SIGNAL_SUPPORT)
        #define DSR_BASE      (*ADR_MSR)
        #define DSR_BITMASK   0x20                       // bit mask to set DSR port pin 
        #define DSR_BIT       (DSR_BASE & DSR_BITMASK) 
        #define DSR_STATE     (DSR_BIT?1:0)              // returns 1 if signal is activ
    #else
        #define DSR_STATE    SIGNAL_NOT_SUPPORTED  // DSR signal not supported
        #define DSR_DPR
    #endif

    #if(DCD_SIGNAL_SUPPORT)
        #define DCD_BASE      (*ADR_MSR)
        #define DCD_BITMASK   0x80                       // bit mask to set DCD port pin 
        #define DCD_BIT       (DCD_BASE & DCD_BITMASK) 
        #define DCD_STATE     (DCD_BIT?1:0)              // returns 1 if signal is activ
    #else
        #define DCD_STATE    SIGNAL_NOT_SUPPORTED  // DCD signal not supported
        #define DCD_DPR
    #endif

    #if(RI_SIGNAL_SUPPORT)
        #define RI_BASE      (*ADR_MSR)
        #define RI_BITMASK   0x40                       // bit mask to set RI port pin 
        #define RI_BIT       (RI_BASE & RI_BITMASK) 
        #define RI_STATE     (RI_BIT?1:0)               // returns 1 if signal is activ
    #else
       #define RI_STATE     SIGNAL_NOT_SUPPORTED   // RI signal not supported
       #define RI_DPR
    #endif

    // RTS wird hier erst mal gelscht. Nach erfolgreichem Abschlu der Initialisierung
    // wird dann das Signal gesetzt. Dadurch soll verhindert werden, das der Sender mit
    // der bertragung beginnt, noch bevor der Empfnger bereit ist. 
    #define RTSCTS_INIT() { CLR_RTS ; CLR_DTR ; }

#else // #if(HW_HANDSHAKE_SUPPORT)
    // Hardware-Handshake wird nicht untersttzt. Um XON-XOFF-Protokoll
    // zu untersttzen, wird das Makro fr CTS verwendet, um 
    // Softwareflags zu steuern.
    // CTS steuert den Sendvorgang. Wurde XOFF empfangen, so geht CTS nach FALSE.
    // Wurde XON empfangen, so geht CTS auf TRUE.
    // fXOFF steuert das Empfangen von Zeichen. Ist der Puffer voll, so wird
    // fXOFF gesetzt und XOFF gesendet. Ist der Puffer leer, so wird XON gesendet.
    static  BOOL fCTS = TRUE;

    #define RTS_STATE   TRUE
    #define CTS_STATE   fCTS
    #define DTR_STATE   SIGNAL_NOT_SUPPORTED
    #define DSR_STATE   SIGNAL_NOT_SUPPORTED
    #define DCD_STATE   SIGNAL_NOT_SUPPORTED
    #define RI_STATE    SIGNAL_NOT_SUPPORTED
    #define CLR_RTS     
    #define SET_RTS     
    #define CLR_CTS     (fCTS = FALSE)
    #define SET_CTS     (fCTS = TRUE)
    #define CLR_DTR
    #define SET_DTR
    #define RTS_SIGNAL_SUPPORT   FALSE
    #define DTR_SIGNAL_SUPPORT   FALSE
    #define RTSCTS_INIT()        { fXOFF_l=FALSE; fXON_l=FALSE; SET_CTS ;}

#endif // #if(HW_HANDSHAKE_SUPPORT)

#define XON  0x11
#define XOFF 0x13



//---------------------------------------------------------------------------
// local types
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// modul globale vars
//---------------------------------------------------------------------------
static BYTE bUseProtocol_l;
static BOOL fXOFF_l;
static BOOL fXON_l;



//---------------------------------------------------------------------------
// local function prototypes
//---------------------------------------------------------------------------

// function prototypes
BOOL SerInit        (tSioCfgParam FAR* SioCfgParam_p);
BOOL SerClose       (void);
BOOL TxFifoPutByte  (char bValue_p);
BOOL RxFifoGetByte  (char FAR* pbValue_p);
BOOL TxFifoDataSize (DWORD FAR* pdwSize_p);
BOOL RxFifoDataSize (DWORD FAR* pdwSize_p);
BOOL GetState       (tSioState FAR* pSioState_p);
BOOL SetState       (tSioState FAR* pSioState_p);

#ifndef _USE_RTOS_
    static void  SioInstallIntHandler   (void);
    static void  SioDeinstallIntHandler (void);
#endif





/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*          C L A S S  <SerialDrv Hardware interface>                      */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
//
// Description:
// Die Klasse stellt Funktionen zur Verfgung fr den Zugriff auf die
// Hardware der seriellen Schnittstelle. Diese Funktionen werden nur intern
// verwendet, um den Zugriff auf die Hardware im Generic-Treiber zu kapseln.
//
// Zeichen senden:
// Das Senden von Zeichen erfolgt ausschlielich in der TX-ISR. Beim Senden 
// von Zeichen mit Hardware-Protokoll wird vor dem bertragen des nchsten 
// Zeichens geprft, ob der Empfnger auch bereit ist, Zeichen zu empfangen. 
// Wurde das Senden gesperrt, so mu zyklisch durch die Applikation der
// Sendekanal getriggert werden, um bei wieder empfangsbereitem Empfnger
// mit der bertragung fortzufahren. Da sowohl das Senden von Zeichen als 
// auch das Stoppen der bertragung innerhalb der TX-ISR realisiert ist, ist
// ein Signalisierung aus zwei Richtungen erforderlich.
// Zum einen erfolgt eine Signalisierung durch die Hardware, wenn ein Zeichen
// vollstndig bertragen wurde und der Sendepuffer leer ist. In diesem 
// Betriebsmodus darf jedoch die ISR nicht durch die Applikation getriggert 
// werden (Die ISR geht davon aus, da das nchste Zeichen im Sendepuffer
// zu hinterlegen ist). Nach vollstndiger bertragung aller Zeichen oder
// bei Unterbrechen der bertragung durch den Empfnger wird das Signalisieren 
// durch die Hardware gesperrt. Der Betriebsmodus wird umgeschaltet, nun
// kann aus der Applikation die ISR signalisiert werden, um den Status der 
// Handshake-Leitungen zu prfen. Beide Betriebsmodies mssen sich stets
// ausschlieen. Ein Umschalten des Betriebsmodus darf nur innerhalb der ISR
// erfolgen.
// 
/***************************************************************************/

//=========================================================================//
//                                                                         //
//          P R I V A T E   D E F I N I T I O N S                          //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
// const defines
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// local types
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// lokale vars
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// local function prototypes
//---------------------------------------------------------------------------

static void  SioDrvIsrHandler   (void);
static void  SioDrvTxIsrHandler (void);
static void  SioDrvRxIsrHandler (void);





//=========================================================================//
//                                                                         //
//          P U B L I C   F U N C T I O N S                                //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
//
// Function:     SioDrvHwInit
//
// Description:  The function configures the serial interface.
//
//
// Parameters:   parameter
//
// 
// Returns:      error code
//
//
// State:        
//
//---------------------------------------------------------------------------
 
static BOOL  SioDrvHwInit (tSioCfgParam FAR* pSioCfgParam_p)
{

DWORD dwBaudValue;
BYTE  bDLL;
BYTE  bDLM;
BYTE  bLCR;
BOOL  fRetcode;


    fRetcode  = TRUE;

    bDLL = bDLM = bLCR = 0;
    
    //------------------------------------------------
    switch (pSioCfgParam_p->m_dwBaudrate)
    {

        case    1200L:  // set baudrate to   1200 baud     
        case    2400L:  // set baudrate to   2400 baud     
        case    9600L:  // set baudrate to   9600 baud
        case   19200L:  // set baudrate to  19200 baud     
        case   38400L:  // set baudrate to  38400 baud     
        case   57600L:  // set baudrate to  57600 baud
        case  115200L:  // set baudrate to 115200 baud    

            // Calculate the devisor latch value for the desired baudrate
            // divisor = UART_CLOCK / (Baudrate * 16)
            // CPU-Frequenz wird mit 10 skaliert, um eine Stelle nach dem Komma zu runden
            dwBaudValue = (UART_CLOCK * 10L) / (16* pSioCfgParam_p->m_dwBaudrate);
            // Um Rundungsfehler so klein mglich zu halten ...
            if ( (dwBaudValue  % 10) >= 5)
               dwBaudValue += 5;
            // Auf ganze Zahlen runden ... 
            dwBaudValue = dwBaudValue/10;
            // ... divisor latch LSB und MSB laden
            bDLL = (BYTE)dwBaudValue;
            bDLM = (BYTE)(dwBaudValue>>8);
        break;

        default:
            // set error code for baudrate not supported   
            pSioCfgParam_p->m_bErrCode = kSioDrvInitBaudrateNotSupported; 
            fRetcode = FALSE; 
            goto Exit;            

    }// End switch
    
    //------------------------------------------------
    switch (pSioCfgParam_p->m_bDataBits)
    {
        case 5:
             bLCR |= 0x00;//set five databits
             break;

        case 6:
             bLCR |= 0x01;//set six databits
             break;

        case 7:
             bLCR |= 0x02;//set seven databits
             break;

        case 8:  
             bLCR |= 0x03;  //set eight databits
             break;

        default:        
            // set error code for databits not supported
            pSioCfgParam_p->m_bErrCode = kSioDrvInitNumOfDataBitsNotSupported; 
            fRetcode = FALSE;                 
            goto Exit;     
   
   }// End switch

   //------------------------------------------------
   // stick parity is not supported
   switch (pSioCfgParam_p->m_bParity)
   {
       
        case PARITY_NONE: 
            bLCR &= ~0x08; //disable parity check
            break;
    
        case PARITY_ODD:
            bLCR &= ~0x10; //enable odd parity
            bLCR |=  0x08; //enable party check
            break;  
    
        case PARITY_EVEN:
            bLCR |= 0x018;  //enable even parity check
            break;
    
        default:
            //set error code for parity not supported
            pSioCfgParam_p->m_bErrCode = kSioDrvInitSelParityNotSupported; 
            fRetcode = FALSE; 
            goto Exit;
    }// End switch

    //------------------------------------------------
    switch (pSioCfgParam_p->m_bStopBits)
    {
        case 1:
            bLCR &= ~0x04;  //set one stop bit
            break;

        case 2:
            bLCR |= 0x04;   //set two stopbits
            break;

        default:
            //set error code for stopbits not supported
            pSioCfgParam_p->m_bErrCode = kSioDrvInitNumOfStopBitsNotSupported; 
            fRetcode = FALSE;
            goto Exit;

    }// End switch

    //------------------------------------------------
    switch (pSioCfgParam_p->m_bProtocol)
    {
        #if (HW_HANDSHAKE_SUPPORT)
          case 2:
              bUseProtocol_l = RTS_CTS_PROTOCOL; // Hardware handshake is used
              break;
        #else
          // kein Protokoll 
          case 0: 
              bUseProtocol_l = NO_PROTOCOL; // no protocol is used           
              break;

          case 1:
              bUseProtocol_l = XON_XOFF_PROTOCOL; // XON/XOFF protocol is used               
              break;
        #endif
     

        default:
            //set error code for protocol not supported
            pSioCfgParam_p->m_bErrCode = kSioDrvInitSelProtocolNotSupported; 
            fRetcode = FALSE; 
            goto Exit;

    }


    //------------------------------------------------
    // Install interrupt handler
    //------------------------------------------------
    // Die Freigabe des MC-Interrupts mu noch vor
    // der Freigabe der Interrupts in der UART erfolgen.
    // Anderenfalls wird Interrupt ausgelst, noch bevor
    // Ereignis durch die CPU registriert werden kann. 
    // Ereignisse werden dadurch nicht weiter verarbeitet.
    #ifdef _USE_RTOS_
    {
        RtosCreateTask (TASK_XUART,      SioDrvIsrHandler);
        RtosCreateTask (TASK_TX_HANDLER, SioDrvTxIsrHandler);
    }
    #else
    {
        SioInstallIntHandler();
    }
    #endif

    //------------------------------------------------
    // Initialize internal interface
    //------------------------------------------------

    // Modem control Register (MCR)
    *ADR_MCR = MCR_REGISTER;

    // Divisor Latch LSB (DLL) and Divisor Latch MSB (DLM)
    // -----> DLAB = 1
    *ADR_LCR = bLCR + DLAB_BIT;
    *ADR_DLL = bDLL;
    *ADR_DLM = bDLM;
    // -----> DLAB = 0, set LCR
    *ADR_LCR = bLCR;

    CLR_TX_FIFO();
    CLR_RX_FIFO();

    // FIFO control Register (FCR) setzen
    *ADR_FCR = FCR_REGISTER;

    // Interrupt Enable Register (IER) setzen     
    *ADR_IER = IER_REGISTER;

Exit:

    return (fRetcode);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvHwClose
//
// Description:  The function disabled the serial interface and 
//               stops the interrupt processing.
//
//
// Parameters:   none
//
// 
// Returns:      TRUE
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvHwClose (void)
{

    // Modem control Register (MCR)
    // Steuerleitungen inaktiv
    *ADR_MCR = 0;

    // Interrupts sperren
    *ADR_IER = 0;



    //------------------------------------------------
    // Deinstall interrupt handler
    //------------------------------------------------
    #ifdef _USE_RTOS_
    {
        RtosDeleteTask (TASK_XUART);
        RtosDeleteTask (TASK_TX_HANDLER);
    }
    #else
    {
        SioDeinstallIntHandler();
    }
    #endif


    return (TRUE);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvPutChar
//
// Description:  The function sends one character.
//
//
// Parameters:   character
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvPutChar (char cChar_p)
{

    *ADR_THR = cChar_p;

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvGetChar
//
// Description:  The function reads one character from serial interface.
//
//
// Parameters:   none
//
// 
// Returns:      character
//
//
// State:        
//
//---------------------------------------------------------------------------

static char  SioDrvGetChar (void)
{

    return ((char) *ADR_RBR);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvGetErrorState
//
// Description:  The function determines the error state of serial interface.
//
//
// Parameters:   none
//
// 
// Returns:      Error flags
//               Bit0: Overrun Error
//               Bit1: Parity Error
//               Bit2: Framing Error
//
//
// State:        
//
//---------------------------------------------------------------------------

static BYTE  SioDrvGetErrorState (void)
{

BYTE bStatus;


    bStatus  = *ADR_LSR;
    // Overrun, Framing and Parity Error filtern
    bStatus &= 0x0E;
    
    return (bStatus>>1);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvSignalTxIsr
//
// Description:  The function sends a signal to tx task.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvSignalTxIsr (void)
{

    #ifdef _USE_RTOS_
    {
        RtosEnableTask(TASK_TX_HANDLER);
    }
    #else
    {
        UART_TX_TASK_BIT_IC_IE = 1;
    }
    #endif

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvDisableTxIsr
//
// Description:  The function disables tx task.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvDisableTxIsr (void)
{

    #ifdef _USE_RTOS_
    {
        RtosDisableTask(TASK_TX_HANDLER);
        RtosSendSignal (TASK_TX_HANDLER);
    }
    #else
    {
        UART_TX_TASK_BIT_IC_IE = 0;
        UART_TX_TASK_BIT_IC_IR = 1;
    }
    #endif

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvIsrHandler
//
// Description:  Interrupt-Handler for Interrupts of UART
//               Function is called by interrupt of
//               UART and switchs according to the interrupt event
//               to the specific Interrupt-Handler.
//
//
// Parameters:   none
// 
//
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvIsrHandler (void)
{

BYTE bIdent;


    for (;;)
    {
        bIdent = (*ADR_IIR);
    
        switch (bIdent & 0x0F)
        {
            //--------------------------------------------------------   
            // Alle Interrupts verarbeitet
            case 1:
                goto Exit;
            break;

            //--------------------------------------------------------   
            case 2: // Transmitter holding register empty
                    // Der Sendeinterrupt wird sowohl durch die HW als
                    // auch Applikation getriggert. Um nicht die  
                    // Ausfhrung ber Semaphoren verriegeln zu 
                    // mssen, wird hier lediglich ein Signal gesendet.
                    // Die Ausfhrung erfolgt dann unmittelbar nach 
                    // Beenden dieses Handlers.
                    #ifdef _USE_RTOS_
                    {
                        RtosSendSignal (TASK_TX_HANDLER);
                    }
                    #else
                    {
                        UART_TX_TASK_BIT_IC_IR = 1;
                    }
                    #endif
            break;
    
            //--------------------------------------------------------   
            case 4:  // Receive Data available
            case 12: // character timeout interrupt
                    SioDrvRxIsrHandler();
            break;
    
            //--------------------------------------------------------   
            case 0: // Modem status interrupt -> not supported
            case 6: // Receiver line status interrupt
            default:
            break;  // not supported
        }

    }


Exit:

    return;

}

/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*          C L A S S  <Serial Driver generic part>                        */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
//
// Description:
// Die Klasse stellt hardwareunabhngig Funktionen des seriellen Treibers 
// zur Verfgung. 
//
/***************************************************************************/


//=========================================================================//
//                                                                         //
//          P R I V A T E   D E F I N I T I O N S                          //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
// const defines
//---------------------------------------------------------------------------

// This macro sets a status bit (x) on global status variable.
#define SETSTAT()      SioDrvSignalTxIsr()
// This macro clears a status bit (x) on global status variable.
#define CLRSTAT()      SioDrvDisableTxIsr()

//---------------------------------------------------------------------------
// local types
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// modul globale vars
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// local function prototypes
//---------------------------------------------------------------------------

static void  SerTriggerTx         (void);
static void  SerTriggerRx         (void);

static void  SioDrvRBufInit       (void);
static void  SioDrvRxRBufClear    (void); 
static void  SioDrvTxRBufClear    (void);
static BOOL  SioDrvRxRBufRead     (BYTE FAR* pbData_p);
static BOOL  SioDrvTxRBufRead     (BYTE FAR* pbData_p);
static BOOL  SioDrvRxRBufWrite    (BYTE bData_p);
static BOOL  SioDrvTxRBufWrite    (BYTE bData_p);
static WORD  SioDrvTxRBufDataSize (void);
static WORD  SioDrvRxRBufDataSize (void);




//=========================================================================//
//                                                                         //
//          P U B L I C   F U N C T I O N S                                //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
//
// Function:     SerInit
//
// Description:  This function is called by application to initialise 
//               serial interface.
//
//
// Parameters:   configuration parameter
//
// 
// Returns:      error code
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  SerInit (tSioCfgParam FAR* pSioCfgParam_p)
{

BOOL fRetcode = TRUE;


    bUseProtocol_l = 0;
    fXOFF_l        = FALSE;
    fXON_l         = FALSE;

    // Sende- und Empfangspuffer init.
    SioDrvRBufInit();

    // Steuerleitungen init., Anfangswerte setzen
    // RTS = 0
    // DTR = 0
    RTSCTS_INIT();

    // per Default wird Echo nicht untersttzt
    pSioCfgParam_p->m_bProtocol &= ~SIODRV_ECHO_SUPPORTED;

    // initialise serial interface
    fRetcode = SioDrvHwInit (pSioCfgParam_p);

    // Initialisieren erfolgreich ?
    if (fRetcode)
    {
       pSioCfgParam_p->m_bErrCode = kSioDrvInitOk;

       #if (SIO_ECHO_SUPPORTED)
       {
          // Der Treiber hinterlegt in m_bProtocol, ob der
          // Echo-Support untersttzt wird.
          pSioCfgParam_p->m_bProtocol |= SIODRV_ECHO_SUPPORTED;
       }
       #endif

       // TX-Task sperren
       CLRSTAT();

       // Initial. des Treibers erfolgreich 
       // abgeschlossen, RTS = 1 setzen.
       SET_RTS;
       fXON_l = TRUE;
       // Die DTR-Leitung bleibt hier auf 0, DTR wird
       // aus der Applikation gesetzt. 
    }    

    return (fRetcode);

}



//---------------------------------------------------------------------------
//
// Function:     SerClose
//
// Description:  This function is called by application to deinitialise 
//               serial interface.
//
//
// Parameters:   none
//
// 
// Returns:      true
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  SerClose (void)
{

    CLR_RTS;
    SioDrvHwClose();

    return (TRUE);

}


//---------------------------------------------------------------------------
//
// Function:     GetState
//
// Description:  The function gets the state of modem control lines.
//
//
// Parameters:   pSioState_p    -pointer to struc tSioState
// 
//
// Returns:      BOOL           -TRUE/FALSE
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  GetState (tSioState FAR* pSioState_p)
{   

DWORD dwSize;


    pSioState_p->m_bErrCode = kSioDrvInitOk;

    //get number of characters in the Rx FIFO
    if(RxFifoDataSize(&dwSize))
    {
        pSioState_p->m_dwCbRxQ = dwSize;
    }// End if
    else
    {
        pSioState_p->m_dwCbRxQ = 0L;    
    }//End else

    //get number of characters in the Tx FIFO
    if(TxFifoDataSize(&dwSize))
    {
        pSioState_p->m_dwCbTxQ = dwSize;
    }//End if
    else
    {
        pSioState_p->m_dwCbTxQ = 0L;    
    }//End else

    pSioState_p->m_dwSoRxQ = SERIAL_RX_FIFO_SIZE;// size of Tx queue
    pSioState_p->m_dwSoTxQ = SERIAL_TX_FIFO_SIZE;// size of Rx queue

    // Hardware handshake and modem control lines
    pSioState_p->m_bCTS = CTS_STATE;
    pSioState_p->m_bDSR = DSR_STATE;
    pSioState_p->m_bDCD = DCD_STATE;
    pSioState_p->m_bRI  = RI_STATE;
    pSioState_p->m_bRTS = RTS_STATE;
    pSioState_p->m_bDTR = DTR_STATE;

    //status register
    pSioState_p->m_wSioStat = (WORD)SioDrvGetErrorState();

    SerTriggerTx();
    SerTriggerRx();

    return (TRUE);   

}



//---------------------------------------------------------------------------
//
// Function:     SetState
//
// Description:  The function sets the state of modem control lines.
//
//               The modem control lines are only effected, if 
//               the the hardware protocol was selected. 
//
//               If software protocol XON/XOFF or no protocoll is used, 
//               the state of RTS and CTS signal can not be changed.
//
// Parameters:   pSioState_p    -Pointer auf Statusstructur
// 
//
// Returns:      TRUE:  sucessfull 
//               FALSE: funktion finishs with error
//                      error code --> pSioState_p->m_bErrCode
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  SetState (tSioState FAR* pSioState_p)
{

BYTE bRetcode = TRUE;


    pSioState_p->m_bErrCode = kSioDrvStatOk;

    // Lschen Empfangspuffer, Sendpuffer ?
    switch (pSioState_p->m_bClr)
    {
        // Empfangspuffer lschen
        case CLR_RX_BUFFER:
            CLR_RX_FIFO();
            SioDrvRxRBufClear();
        break;

        // Sendepuffer lschen
        case CLR_TX_BUFFER:
            CLR_TX_FIFO();
            SioDrvTxRBufClear();
        break;

        // Empfangspuffer und Sendepuffer lschen
        case CLR_RX_TX_BUFFER:
            CLR_RX_FIFO();
            SioDrvRxRBufClear();
            CLR_TX_FIFO();
            SioDrvTxRBufClear();
        break;

        default:
        break;  
    }

    //---------------------------------------------------------
    // If hardware handshake is used, the state of the RTS
    // and DTR signal can be set by software.
    if (bUseProtocol_l == RTS_CTS_PROTOCOL)
    {
        // RTS-Signal-Support was configured?
        #if(RTS_SIGNAL_SUPPORT)
        {
            switch (pSioState_p->m_bRTS)
            {
                case SIGNAL_SET_INACTIVE:
    
                    CLR_RTS;// clear RTS signal
                    break;
        
                case SIGNAL_SET_ACTIVE:
        
                    SET_RTS;// set RTS signal
                    break;  
    
                case SIGNAL_NOT_CHANGE: 
                    break;
    
            }// End switch                                    

        }// End #if 
        
        #else
        {   // RTS-Signal-Support was not set, return an Error.
            pSioState_p->m_bErrCode = kSioDrvStatRtsSetError;
            bRetcode = FALSE;
        }// End #else
        #endif
    
        //---------------------------------------------------
        // DTR-Signal-Support was configured?
        #if(DTR_SIGNAL_SUPPORT)
        {
            switch (pSioState_p->m_bDTR)
            {
                case SIGNAL_SET_INACTIVE:
    
                       CLR_DTR;// clear DTR signal
                       break;
        
                case SIGNAL_SET_ACTIVE:
    
                       SET_DTR;// set DTR signal
                       break;  
    
                case SIGNAL_NOT_CHANGE: break;
                
    
            }// End switch                                    
        }
        #else
        {   // DTR-Signal-Support was not set, return an Error.
            pSioState_p->m_bErrCode = kSioDrvStatDtrSetError;
            bRetcode = FALSE;
        }
        #endif     

        goto Exit;
    }//End if       

    //---------------------------------------------------------
    // If software protocol XON/XOFF or no protocoll is used, the state of
    // RTS and CTS signal can not be changed.
    if ((bUseProtocol_l == XON_XOFF_PROTOCOL) || (bUseProtocol_l == NO_PROTOCOL))
    {
        if (pSioState_p->m_bRTS != SIGNAL_NOT_CHANGE)
        {
            pSioState_p->m_bErrCode = kSioDrvStatRtsNotSupported;
            bRetcode = FALSE;
            goto Exit;
        }// End if
    
        if (pSioState_p->m_bDTR != SIGNAL_NOT_CHANGE)
        {
            pSioState_p->m_bErrCode = kSioDrvStatDtrNotSupported;
            bRetcode = FALSE;
            goto Exit;
        }// End if
        
        // The actual signal state of RTS or DTR not change
        else
        {   //no error
            goto Exit;
        }
    }// End if


Exit:

    return (bRetcode);   

}



//---------------------------------------------------------------------------
//
// Function:     TxFifoPutByte
//
// Description:  This function is called by application to put
//               a character to FIFO of serial interface. After
//               character has been written, the transmission is triggered.
//
//
// Parameters:   character
//
// 
// Returns:      Errorcode
//               TRUE:  character sent
//               FALSE: character not sent
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  TxFifoPutByte (char bChar_p)
{

BOOL fRes;
BOOL fBreak;



    // Unabhngig vom bertragungsmode wird stets
    // das Zeichen im Sendepuffer abgelegt und die
    // TX-Task freigegeben. 
    fRes = SioDrvTxRBufWrite (bChar_p);

    do 
    {
       // TX-Task freigegeben, um Sendung auszufhren
       SerTriggerTx();

       //------------------------------------------------------------
       // Sendefunktion nicht blockierend
       #if USE_BLOCKED_MODE==FALSE
           fBreak = FALSE;
       #else
       //------------------------------------------------------------
       // Im blockierenden Mode wird hier gewartet,
       // bis das Zeichen bertragen wurde.
           fBreak = (SioDrvTxRBufDataSize() != 0)?TRUE:FALSE;
       #endif

    }
    while (fBreak);

    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     RxFifoGetByte
//
// Description:  This function is called by application to get
//               a character from FIFO of serial interface. After
//               character has been read, the receive interrupt chain
//               is triggered again.
//
//
// Parameters:   pointer to memory for storing character
//
// 
// Returns:      error code
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  RxFifoGetByte (char FAR* pbValue_p)
{
    // Es wird versucht ein Zeichen zu lesen. Ist 
    // der Puffer leer, so wird durch Setzen der Steuerleitung RTS
    // dem Sender signalisiert, das weitere Zeichen empfangen werden knnen. 
    if ( !SioDrvRxRBufRead(pbValue_p) )
    {
        // Die Funktion prft, ob gegebenenfalls das
        // Handshake RTS oder XON zu senden ist.
        SerTriggerRx();

        return (FALSE);
    }

    return (TRUE);

}



//---------------------------------------------------------------------------
//
// Function:     TxFifoDataSize
//
// Description:  This function is called by application to get
//               the size of data in TX-FIFO of serial interface. 
//
//
// Parameters:   pointer to memory for storing size
//
// 
// Returns:      error code
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  TxFifoDataSize(DWORD FAR* pdwSize_p)
{

    *pdwSize_p = SioDrvTxRBufDataSize();

    return (TRUE);

}



//---------------------------------------------------------------------------
//
// Function:     RxFifoDataSize
//
// Description:  This function is called by application to get
//               the size of data in RX-FIFO of serial interface. 
//
//
// Parameters:   pointer to memory for storing size
//
// 
// Returns:      error code
//
//
// State:        
//
//---------------------------------------------------------------------------

BOOL  PUBLIC  RxFifoDataSize(DWORD FAR* pdwSize_p)
{

    *pdwSize_p = SioDrvRxRBufDataSize();

    return (TRUE);

}





//=========================================================================//
//                                                                         //
//          P R I V A T E   F U N C T I O N S                              //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
//
// Function:     SioDrvTxIsrHandler
//
// Description:  Interrupt-Handler for tx channel
//
//
// Parameters:   none
// 
//
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------
//
// Erstes Zeichen senden: Sende-Task nicht aktiv, Sendefunktion gibt Sende-Task frei
// Zeichen senden: ISR prft, ob Zeichen vorhanden und sendet diese. Ist Sendepuffer
// leer, dann Sende-Task sperren.
// Receiver nicht bereit: Senden kann nicht erfolgen, da Empfnger nicht bereit.
// Die Sendetask wird gesperrt. Innerhalb der Applikation mu zyklisch eine Funktion 
// gerufen werden, die die Sendetask frei gibt, um ein fortfahren der Senung zu prfen.

static void  SioDrvTxIsrHandler (void)
{

char bTxChar;


    // HW-Handshake: check if receiver is ready
    // XON-XOFF:     check if state == XOFF
    if ( !CTS_STATE )
    {
       // Der Empfnger ist nicht bereit, weitere Zeichen
       // zu empfangen. Die TX-Task bleibt im Running, wird 
       // aber zurckgestellt, bis ein Ereignis eintrifft CTS = TRUE). 
       CLRSTAT();
       return;
    }

    // XON-XOFF: check to send XOFF requested
    if ( fXOFF_l == TRUE )
    {
       // Der Sender mu ein XOFF einfgen, um das weitere
       // Senden der Gegenstelle zu unterbrechen.
       SioDrvPutChar(XOFF);
       // RTS wieder auf TRUE setzen, um zu verhindern,
       // das stndig XOFF gesendet wird. fXON auf TRUE
       // Setzen um zu signalisieren, da bei leerem Puffer
       // XON gesendet werden kann.
       fXOFF_l = FALSE; 
       fXON_l  = TRUE;

       // Die TX-Task bleibt weiter aktiv, da ja nur in den 
       // laufenden Sendestrom ein XOFF eingefgt wurde.
       SETSTAT();
       return;
    }

    // Sind Zeichen im Sendepuffer?
    if (SioDrvTxRBufRead(&bTxChar))
    {
       // Zeichen sind im Sendepuffer vorhanden.
       SioDrvPutChar(bTxChar);
       // Die TX-Task bleibt weiter aktiv.
       SETSTAT();
    }
    else
    {   
       // Der Sendepuffer ist leer, daher die Tx-Task 
       // sperren.
       CLRSTAT();
    }

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvRxIsrHandler
//
// Description:  Interrupt-Handler for rx channel
//
//
// Parameters:   none
// 
//
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvRxIsrHandler (void)
{

char bRxChar;


    bRxChar = SioDrvGetChar();

    // XON-XOFF-Steuerung beachten?
    if (bUseProtocol_l == XON_XOFF_PROTOCOL)
    {
       if (bRxChar == XOFF)
       {
          // Zeichen nicht im Empfangspuffer ablegen,
          // nur Status sichern, um weiteren Sendeverlauf zu steuern
          CLR_CTS;
          return;
       }

       if (bRxChar == XON)
       {
          // Zeichen nicht Empfangspuffer ablegen,
          // nur Status sichern, um weiteren Sendeverlauf zu steuern
          SET_CTS;
          return;
       }     
    }

    // put received byte to FIFO
    SioDrvRxRBufWrite(bRxChar);
    
    // get byte from register
    if (SioDrvRxRBufDataSize() > SER_RXFIFO_STOP_LEVEL)
    {
        // HW-Handshake: RTS lschen und damit Sender sperren
        if (bUseProtocol_l == RTS_CTS_PROTOCOL)
        {
           CLR_RTS;
        }

        // XON-XOFF: Flag setzen, um XOFF in der Sende-Task zu senden
        if (bUseProtocol_l == XON_XOFF_PROTOCOL)
        {
           fXOFF_l = TRUE;
           // Ist die Sende-Task gerade nicht aktiv, so 
           // Task signalisieren, um die bertragung zu starten.
           SerTriggerTx();
        }
    }

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvErrIsrHandler
//
// Description:  Interrupt Handler for Error-Interrupt
//
//
// Parameters:   none
// 
//
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------
/*
static void  SioDrvErrIsrHandler (void)
{

}
*/


//---------------------------------------------------------------------------
//
// Function:     SerTriggerRx
//
// Description:  This function is called by serial driver to trigger
//               data receive just do handshaking.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SerTriggerRx (void)
{

    if (SioDrvRxRBufDataSize() == 0)
    {
        if (bUseProtocol_l == RTS_CTS_PROTOCOL)
        {
           SET_RTS;        // weitere Zeichen anfordern, da Puffer leer
        }
        // XON senden?
        if (fXON_l  && (bUseProtocol_l == XON_XOFF_PROTOCOL))
        {
           fXON_l = FALSE;
           TxFifoPutByte(XON);
        }
    }

}



//---------------------------------------------------------------------------
//
// Function:     SerTriggerTx
//
// Description:  This function is called by serial driver to trigger
//               data sending.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SerTriggerTx(void)
{
    
    // Die TX-Task freigeben.
    SETSTAT();

}





/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*          C L A S S  <Ring buffer>                                       */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
//
// Description:
// Die Ringpuffer-Verwaltung ist so implementiert, da fr den Zugriff
// auf den Puffer keine Sperre fr den komplementren Zugriff auf den Puffer
// aus einer hherpriorisierten Task notwendig ist. D.h. Ein Lesezugriff
// kann durchaus durch einen Schreibzugriff unterbrochen werden. Es ist 
// zu verhindern, da ein Lesezugriff durch einen Lesezugriff auf den 
// gleichen Puffer aus einer hherpriorisierten Task erfolgt. Fr die 
// korrekte Funktion ist weiterhin wichtig, da das Lesen oder Schreiben eines
// der Puffer-Indizes stets in einer nichtunterbrechbaren Sequenz erfolgt
// (Ausfhrung in einem Maschinenbefehl). Da das von Maschine zu Maschine
// unterschiedlich ist, wurden die Datentypen fr die Indizes ber Konstanten
// definiert, um maschinenabhngig den Zugriff steuern zu knnen und zustzlich 
// (vorsorglich, falls das aus technischen Grnden nicht anders mglich ist, da
// z.B. der Puffer sehr gro sein mu) der Zugriff durch eine ATOMIC/ENDATOMIC
// Anweisung eingerahmt. Diese Makros knnen durch Funktionen oder Anweisungen
// ersetzt werden, die dann whrend des Zugriffs die hherpriorisierte Task 
// sperren.
// 
//
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// const defines
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//  Target specific definitions
//---------------------------------------------------------------------------

#ifdef __C166__
    #define ATOMIC_START() _atomic_(0)
    #define ATOMIC_END()   _endatomic_() 
#else
    #define ATOMIC_START() 
    #define ATOMIC_END()   
#endif



//---------------------------------------------------------------------------
// local types
//---------------------------------------------------------------------------
// Der Type fr den Buffer-Index mu machinenabhngig so gewhlt
// werden, da das Lesen oder Schreiben des Index nicht durch einen 
// Interrupt unterbrochen werden kann. Beim einem 16Bit-Controller 
// ist das in der Regel ein 16Bit-Zugriff, beim PC ab x386 ein 32Bit-
// Zugriff. Wird auf Grund der Puffergre einer grerer Typ bentigt,
// so sind die ATOMIC-Makros so anzupassen, da eine Unterbrechung
// whrend des Zugriffs nicht mglich ist.
#define tBufIdx WORD



//---------------------------------------------------------------------------
// lokale vars
//---------------------------------------------------------------------------

static BYTE     abRxRingBuff_l[SERIAL_RX_FIFO_SIZE];
static BYTE     abTxRingBuff_l[SERIAL_TX_FIFO_SIZE];
static tBufIdx  wRxBuffWrIdx_l;
static tBufIdx  wRxBuffRdIdx_l;
static tBufIdx  wTxBuffWrIdx_l;
static tBufIdx  wTxBuffRdIdx_l;


//---------------------------------------------------------------------------
// local function prototypes
//---------------------------------------------------------------------------

static BOOL SioDrvTxRBufIsEmpty  (void);
static BOOL SioDrvRxRBufIsEmpty  (void);


//=========================================================================//
//                                                                         //
//          P U B L I C   F U N C T I O N S                                //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
//
// Function:     SioDrvRBufInit
//
// Description:  The function initializes the rx and tx buffer.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvRBufInit (void)
{            

    SioDrvRxRBufClear ();
    SioDrvTxRBufClear ();

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvRxBufClear
//
// Description:  The function clears the rx buffer.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvRxRBufClear (void)
{

    ATOMIC_START();
    {
        wRxBuffWrIdx_l = 0;
        wRxBuffRdIdx_l = 0;
    }
    ATOMIC_END();

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvTxBufClear
//
// Description:  The function clears the tx buffer.
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDrvTxRBufClear (void)
{

    ATOMIC_START();
    {
        wTxBuffWrIdx_l = 0;
        wTxBuffRdIdx_l = 0;
    }
    ATOMIC_END();

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvTxRBufWrite
//
// Description:  The function writes data to tx buffer.
//
//
// Parameters:   data byte
//
// 
// Returns:      error code
//               TRUE:  character has been written
//               FALSE: buffer full
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvTxRBufWrite (BYTE bData_p)
{

WORD  wNextBuffWrIdx;
WORD  wBuffRdIdx;
BOOL  fRes;


    // get buffer index for next write operation
    ATOMIC_START();
    {
        wNextBuffWrIdx = wTxBuffWrIdx_l;
    }
    ATOMIC_END();

    
    wNextBuffWrIdx += 1;
    if (wNextBuffWrIdx >= sizeof(abTxRingBuff_l))
    {
        wNextBuffWrIdx = 0;
    }


    // free buffer availabale (write index and read index not identical)?
    ATOMIC_START();
    {
        wBuffRdIdx = wTxBuffRdIdx_l;
    }
    ATOMIC_END();


    fRes = (wNextBuffWrIdx != wBuffRdIdx);

    if (fRes)
    {
        // free buffer available -> write data to buffer
        abTxRingBuff_l[wTxBuffWrIdx_l] = bData_p;

        ATOMIC_START();
        {
            wTxBuffWrIdx_l = wNextBuffWrIdx;
        }
        ATOMIC_END();
    }


    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvRxRBufWrite
//
// Description:  The function writes data to rx buffer.
//
//
// Parameters:   data byte
//
// 
// Returns:      error code
//               TRUE:  character has been written
//               FALSE: buffer full
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvRxRBufWrite (BYTE bData_p)
{

WORD  wNextBuffWrIdx;
WORD  wBuffRrIdx;
BOOL  fRes;


    // get buffer index for next write operation
    ATOMIC_START();
    {
        wNextBuffWrIdx = wRxBuffWrIdx_l;
    }
    ATOMIC_END();


    wNextBuffWrIdx += 1;
    if (wNextBuffWrIdx >= sizeof(abRxRingBuff_l))
    {
        wNextBuffWrIdx = 0;
    }


    // free buffer availabale (write index and read index not identical)?
    ATOMIC_START();
    {
        wBuffRrIdx = wRxBuffRdIdx_l;
    }
    ATOMIC_END();


    fRes = (wNextBuffWrIdx != wBuffRrIdx);

    if (fRes)
    {
        // free buffer availabale -> write data to buffer
        abRxRingBuff_l[wRxBuffWrIdx_l] = bData_p;

        ATOMIC_START();
        {
            wRxBuffWrIdx_l = wNextBuffWrIdx;
        }
        ATOMIC_END();
    }


    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvTxRBufRead
//
// Description:  The function reads data from tx buffer.
//
//
// Parameters:   pointer to storage data byte
//
// 
// Returns:      error code
//               TRUE:  character has been read
//               FALSE: buffer empty
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvTxRBufRead (BYTE FAR* pbData_p)
{

WORD  wNextBuffRdIdx;
BOOL  fRes;


    // Is at least one data entry available?
    fRes = SioDrvTxRBufIsEmpty();

    if ( fRes )
    {
        // return the data from buffer
        *pbData_p = abTxRingBuff_l[wTxBuffRdIdx_l];


        // get buffer index for next read operation
        ATOMIC_START();
        {
            wNextBuffRdIdx = wTxBuffRdIdx_l;
        }
        ATOMIC_END();


        wNextBuffRdIdx += 1;
        if (wNextBuffRdIdx >= sizeof(abTxRingBuff_l))
        {
            wNextBuffRdIdx = 0;
        }


        ATOMIC_START();
        {
            wTxBuffRdIdx_l = wNextBuffRdIdx;
        }
        ATOMIC_END();
    }

    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvRxRBufRead
//
// Description:  Function reads data from the Rx-Buffer.
//
//
// Parameters:   pointer to storage data byte
//
// 
// Returns:      Errorcode
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL SioDrvRxRBufRead (BYTE FAR* pbData_p)
{

WORD  wNextBuffRdIdx;
BOOL  fRes;


    // Is at least one data entry available?
    fRes = SioDrvRxRBufIsEmpty();

    // buffer is not empty
    if ( fRes )
    {
        // return the data from buffer
        *pbData_p = abRxRingBuff_l[wRxBuffRdIdx_l];


        // get buffer index for next read operation
        ATOMIC_START();
        {
            wNextBuffRdIdx = wRxBuffRdIdx_l;
        }
        ATOMIC_END();
        

        wNextBuffRdIdx += 1;
        if (wNextBuffRdIdx >= sizeof(abRxRingBuff_l))
        {
            wNextBuffRdIdx = 0;
        }


        ATOMIC_START();
        {
            wRxBuffRdIdx_l = wNextBuffRdIdx;
        }
        ATOMIC_END();
    }


    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvRxRBufDataSize
//
// Description:  Function gets the number of data bytes in Rx-Buffer.
//
//
// Parameters:   none
//
// 
// Returns:      data size of rx buffer
//
//
// State:        
//
//---------------------------------------------------------------------------

static WORD  SioDrvRxRBufDataSize (void)
{

WORD  wIdxDiff;


    // get the difference of write index and read index (in a safe section!)
    ATOMIC_START();
    {
        wIdxDiff = wRxBuffWrIdx_l - wRxBuffRdIdx_l;
    }
    ATOMIC_END();


    return (wIdxDiff);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvTxRBufDataSize
//
// Description:  Function gets the number of data bytes in Tx-Buffer.
//
//
// Parameters:   none
//
// 
// Returns:      data size of tx buffer
//
//
// State:        
//
//---------------------------------------------------------------------------

static WORD SioDrvTxRBufDataSize (void)
{

WORD  wIdxDiff;


    // get the difference of write index and read index (in a safe section!)
    ATOMIC_START();
    {
        wIdxDiff = wTxBuffWrIdx_l - wTxBuffRdIdx_l;
    }
    ATOMIC_END();


    return (wIdxDiff);

}





//=========================================================================//
//                                                                         //
//          P R I V A T E   F U N C T I O N S                              //
//                                                                         //
//=========================================================================//

//---------------------------------------------------------------------------
//
// Function:     SioDrvRxRBufIsEmpty
//
// Description:  Function checks if any data is available in Rx-Buffer.
//
//
// Parameters:   none
//
// 
// Returns:      TRUE:  data available
//               FALSE: no data available
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvRxRBufIsEmpty (void)
{

BOOL  fRes;

    // is at least one data entry available
    // (are write index and read index identical)?
    fRes = (SioDrvRxRBufDataSize() == 0)?FALSE:TRUE;

    return (fRes);

}



//---------------------------------------------------------------------------
//
// Function:     SioDrvTxRBufIsEmpty
//
// Description:  Function checks if any data is available in Tx-Buffer.
//
//
// Parameters:   none
//
// 
// Returns:      TRUE:  data available
//               FALSE: no data available
//
//
// State:        
//
//---------------------------------------------------------------------------

static BOOL  SioDrvTxRBufIsEmpty (void)
{

BOOL  fRes;


    // is at least one data entry available
    // (are write index and read index identical)?
    fRes = (SioDrvTxRBufDataSize() == 0)?FALSE:TRUE;

    return (fRes);

}





/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*          C L A S S  <DirectIntHandling>                                 */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

//=========================================================================//
//                                                                         //
//          P U B L I C   F U N C T I O N S                                //
//                                                                         //
//=========================================================================//

#ifndef _USE_RTOS_

//---------------------------------------------------------------------------
//
// Function:     SioInstallIntHandler
//
// Description:  Function installs and enables interrupt handlers for
//               UART hardware interrupt and TxTask
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioInstallIntHandler (void)
{

/*
    // Set CPU security level to unprotected mode
    SCUSLC = 0xAAAA;
    SCUSLC = 0x5554;
    SCUSLC = 0x96FF;
    SCUSLC = 0;
*/

    // configure uart hardware interrupt (hardware interrupt of uart chip)
    UART_HW_INT_REG_IC = 0;

    UART_HW_INT_REG_TM &= ~UART_HW_INT_MASK_TM;       // clear MOD bits, keep all other
    UART_HW_INT_REG_TM |= UART_HW_INT_MODE_TM;        // set MOD, capture on rising edge

    UART_HW_INT_REG_IC = (UART_HW_INT_GRPXLVL<<8) | (UART_HW_INT_INTLVL<<2) | UART_HW_INT_GRPLVL | 0x40;


    // configure uart tx handler task interrupt (pure software interrupt)
    UART_TX_TASK_REG_IC = (UART_TX_TASK_GRPXLVL<<8) | (UART_TX_TASK_INTLVL<<2) | UART_TX_TASK_GRPLVL | 0x40;

/*
    // Set CPU security level to protected mode
    SCUSLC = 0xAAAA;
    SCUSLC = 0x5554;
    SCUSLC = 0x96FF;
    SCUSLC = 0x1800;
*/

}



//---------------------------------------------------------------------------
//
// Function:     SioDeinstallIntHandler
//
// Description:  Function deinstalls and disables interrupt handlers for
//               UART hardware interrupt and TxTask
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioDeinstallIntHandler (void)
{

/*
    // Set CPU security level to unprotected mode
    SCUSLC = 0xAAAA;
    SCUSLC = 0x5554;
    SCUSLC = 0x96FF;
    SCUSLC = 0;
*/

    // disable uart hardware interrupt (hardware interrupt of uart chip)
    UART_HW_INT_REG_IC  = 0;
    UART_HW_INT_REG_TM &= UART_HW_INT_MASK_TM;        // clear MOD bits, keep all other

    // disable uart tx handler task interrupt (pure software interrupt)
    UART_TX_TASK_REG_IC = 0;

/*
    // Set CPU security level to protected mode
    SCUSLC = 0xAAAA;
    SCUSLC = 0x5554;
    SCUSLC = 0x96FF;
    SCUSLC = 0x1800;
*/

}



//---------------------------------------------------------------------------
//
// Function:     SioInterruptHandler
//
// Description:  ISR for UART hardware interrupt
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioInterruptHandler (void) interrupt UART_HW_INT_NUM //using SioInterruptHandler_RegBank
{

    SioDrvIsrHandler();

}



//---------------------------------------------------------------------------
//
// Function:     SioTxTaskHandler
//
// Description:  ISR for tx handler interrupt
//
//
// Parameters:   none
//
// 
// Returns:      none
//
//
// State:        
//
//---------------------------------------------------------------------------

static void  SioTxTaskHandler (void) interrupt UART_TX_TASK_NUM //using SioTxTaskHandler_RegBank
{

    SioDrvTxIsrHandler();

}


#endif





//EOF
